from traitlets import observe

from jdaviz.core.template_mixin import ViewerSelectMixin
from jdaviz.configs.cubeviz.plugins import Slice
from jdaviz.core.registries import tray_registry

from lcviz.events import EphemerisChangedMessage
from lcviz.viewers import CubeView, TimeScatterView, PhaseScatterView

__all__ = ['TimeSelector']


@tray_registry('time-selector', label="Time Selector", category='app:options')
class TimeSelector(Slice, ViewerSelectMixin):
    """
    See the :ref:`Time Selector Plugin Documentation <time-selector>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``value``  Time of the indicator.  When setting this directly, it will
      update automatically to the value corresponding to the nearest slice, if ``snap_to_slice`` is
      enabled and a cube is loaded.
    * ``show_indicator``
      Whether to show indicator in spectral viewer when slice tool is inactive.
    * ``show_value``
      Whether to show slice value in label to right of indicator.
    * ``snap_to_slice``
      Whether the indicator (and ``value``) should snap to the value of the nearest slice in the
      cube (if one exists).
    """
    _cube_viewer_cls = CubeView
    _cube_viewer_default_label = 'image'

    def __init__(self, *args, **kwargs):
        """

        """
        super().__init__(*args, **kwargs)
        self.plugin_key = "Time Selector"
        self._plugin_description = 'Select active time across all viewers.'
        self.docs_description = "Select time to sync across all viewers (as an indicator in all time/phase viewers or to select the active slice in any image/cube viewers).  The slice can also be changed interactively in any time viewer by activating the slice tool."  # noqa
        self.value_label = 'Time'
        self.value_unit = 'd'
        self.allow_disable_snapping = True

        self.session.hub.subscribe(self, EphemerisChangedMessage,
                                   handler=self._on_ephemeris_changed)

        self.viewer.add_filter(lambda viewer: isinstance(viewer, (TimeScatterView, PhaseScatterView, CubeView)))  # noqa
        self._set_relevant()

    @observe('viewer_items')
    def _set_relevant(self, *args):
        if not len(self.viewer_items):
            self.irrelevant_msg = 'No valid viewers'
        else:
            self.irrelevant_msg = ''

    @observe('vdocs')
    def _update_docs_link(self, *args):
        self.docs_link = f"https://lcviz.readthedocs.io/en/{self.vdocs}/plugins.html#time-selector"

    @property
    def slice_display_unit_name(self):
        # global display unit "axis" corresponding to the slice axis
        return 'time'

    @property
    def valid_slice_att_names(self):
        return ["time", "dt"]

    @property
    def user_api(self):
        api = super().user_api
        # can be removed after deprecated upstream attributes for wavelength/wavelength_value
        # are removed in the lowest supported version of jdaviz
        api._expose = [e for e in api._expose if e not in ('slice', 'wavelength',
                                                           'wavelength_value', 'show_wavelength')]
        return api

    def _on_select_slice_message(self, msg):
        viewer = msg.sender.viewer
        if isinstance(viewer, PhaseScatterView):
            prev_phase = viewer.times_to_phases(self.value)
            new_phase = msg.value
            self.value = self.value + (new_phase - prev_phase) * viewer.ephemeris.get('period', 1.0)
        else:
            super()._on_select_slice_message(msg)

    def _on_ephemeris_changed(self, msg):
        for viewer in self.slice_indicator_viewers:
            if not isinstance(viewer, PhaseScatterView):
                continue
            if viewer._ephemeris_component != msg.ephemeris_label:
                continue
            viewer._set_slice_indicator_value(self.value)



from astropy import units as u
import numpy as np

from jdaviz.core.marks import PluginLine, PluginScatter, SliceIndicatorMarks
from lcviz.viewers import PhaseScatterView

__all__ = ['LivePreviewTrend', 'LivePreviewFlattened', 'LivePreviewBinning']


def _slice_indicator_get_slice_axis(self, data):
    if hasattr(data, 'time'):
        return data.time.value * u.d
    return [] * u.dimensionless_unscaled


SliceIndicatorMarks._get_slice_axis = _slice_indicator_get_slice_axis


class WithoutPhaseSupport:
    def update_ty(self, times, y):
        self.times = np.asarray(times)
        self.x = self.times
        self.y = np.asarray(y)


class WithPhaseSupport(WithoutPhaseSupport):
    def update_ty(self, times, y):
        self.times = np.asarray(times)
        self.update_phase_folding()
        self.y = np.asarray(y)

    def update_phase_folding(self):
        if not hasattr(self, 'times'):
            # update_ty has not been called yet, so there is
            # nothing to phase-fold
            return
        if isinstance(self.viewer, PhaseScatterView):
            self.x = self.viewer.times_to_phases(self.times)
        else:
            self.x = self.times


class LivePreviewTrend(PluginLine, WithoutPhaseSupport):
    def __init__(self, viewer, *args, **kwargs):
        self.viewer = viewer
        super().__init__(viewer, *args, **kwargs)


class LivePreviewFlattened(PluginScatter, WithPhaseSupport):
    def __init__(self, viewer, **kwargs):
        self.viewer = viewer
        kwargs.setdefault('default_size', 16)
        super().__init__(viewer, **kwargs)


class LivePreviewBinning(PluginScatter, WithPhaseSupport):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('default_size', 16)
        super().__init__(*args, **kwargs)



import numpy as np

from glue.viewers.scatter.state import ScatterViewerState

__all__ = ['ScatterViewerState']


class ScatterViewerState(ScatterViewerState):
    def _reset_att_limits(self, ax):
        # override glue's _reset_x/y_limits to account for all layers,
        # not just reference data
        att = f'{ax}_att'
        if getattr(self, att) is None:  # pragma: no cover
            return

        ax_min, ax_max = np.inf, -np.inf
        for layer in self.layers:
            ax_data = layer.layer.data.get_data(getattr(self, att))
            if len(ax_data) > 0:
                ax_min = min(ax_min, np.nanmin(ax_data))
                ax_max = max(ax_max, np.nanmax(ax_data))

        if not np.all(np.isfinite([ax_min, ax_max])):  # pragma: no cover
            return

        lim_helper = getattr(self, f'{ax}_lim_helper')
        lim_helper.lower = ax_min
        lim_helper.upper = ax_max
        lim_helper.update_values()

    def _reset_x_limits(self, *event):
        self._reset_att_limits('x')

    def _reset_y_limits(self, *event):
        self._reset_att_limits('y')

    def reset_limits(self, *event):
        x_min, x_max = np.inf, -np.inf
        y_min, y_max = np.inf, -np.inf

        for layer in self.layers:
            if not layer.visible:  # pragma: no cover
                continue

            x_data = layer.layer.data.get_data(self.x_att)
            y_data = layer.layer.data.get_data(self.y_att)

            x_min = min(x_min, np.nanmin(x_data))
            x_max = max(x_max, np.nanmax(x_data))
            y_min = min(y_min, np.nanmin(y_data))
            y_max = max(y_max, np.nanmax(y_data))

        x_lim_helper = getattr(self, 'x_lim_helper')
        x_lim_helper.lower = x_min
        x_lim_helper.upper = x_max

        y_lim_helper = getattr(self, 'y_lim_helper')
        y_lim_helper.lower = y_min
        y_lim_helper.upper = y_max

        x_lim_helper.update_values()
        y_lim_helper.update_values()

        self._adjust_limits_aspect()



import os
from jdaviz.core.tools import SidebarShortcutPlotOptions, SidebarShortcutExportPlot


ICON_DIR = os.path.normpath(os.path.join(os.path.dirname(__file__), 'data', 'icons'))

# point to the lcviz-version of plot options instead of jdaviz's
SidebarShortcutPlotOptions.plugin_name = 'lcviz-plot-options'
SidebarShortcutExportPlot.plugin_name = 'lcviz-export'


__all__ = []



from numpy.testing import assert_allclose


def test_docs_snippets(helper, light_curve_like_kepler_quarter):
    lcviz, lc = helper, light_curve_like_kepler_quarter

    lcviz.load(lc)
    # lcviz.show()

    flux_col = lcviz.plugins['Flux Column']
    print(flux_col.flux_column.choices)
    # NOTE: choices for docs example are: ['sap_flux', 'sap_bkg', 'pdcsap_flux']
    flux_col.flux_column = 'flux_alt'


def test_plugin_flux_column(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)

    fo = helper.plugins['Flux Column']
    assert len(fo.flux_column.choices) == 2
    assert fo.flux_column.selected == 'flux:orig'

    lc = helper.get_data()
    assert lc.meta.get('FLUX_ORIGIN') == 'flux:orig'
    assert_allclose(lc['flux'], fo._obj.dataset.selected_dc_item['flux:orig'])

    fo.flux_column = 'flux_alt'
    lc = helper.get_data()
    assert lc.meta.get('FLUX_ORIGIN') == 'flux_alt'
    assert_allclose(lc['flux'], fo._obj.dataset.selected_dc_item['flux_alt'])



import numpy as np
from numpy.testing import assert_allclose

from lcviz.marks import LivePreviewTrend, LivePreviewFlattened


def _get_marks_from_viewer(viewer, cls=(LivePreviewTrend, LivePreviewFlattened),
                           include_not_visible=False):
    return [m for m in viewer.figure.marks if isinstance(m, cls)
            if include_not_visible or m.visible]


def test_docs_snippets(helper, light_curve_like_kepler_quarter):
    lcviz, lc = helper, light_curve_like_kepler_quarter

    lcviz.load(lc)
    # lcviz.show()

    flatten = lcviz.plugins['Flatten']
    flatten.open_in_tray()
    flatten.polyorder = 4
    flattened_lc = flatten.flatten(add_data=True)
    print(flattened_lc)


def test_plugin_flatten(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)
    tv = helper.default_time_viewer._obj

    ephem = helper.plugins['Ephemeris']
    pv = ephem.create_phase_viewer()._obj
    f = helper.plugins['Flatten']

    # no marks until plugin opened/active
    assert len(_get_marks_from_viewer(tv)) == 0
    assert len(_get_marks_from_viewer(pv)) == 0

    with f.as_active():
        assert len(_get_marks_from_viewer(tv)) == 2
        assert len(_get_marks_from_viewer(pv)) == 1

        # update period which should update phasing in phase-viewer
        ephem.period = 1.2

        before_polyorder = f.polyorder
        before_update = _get_marks_from_viewer(tv)[0].y

        # test error handling in live-preview
        f.polyorder = -1
        assert f._obj.flatten_err != ''
        assert len(_get_marks_from_viewer(tv)) == 0
        assert len(_get_marks_from_viewer(pv)) == 0

        # update polyorder (live-preview should re-appear and have changed from before)
        f.polyorder = before_polyorder + 1
        assert f._obj.flatten_err == ''
        marks = _get_marks_from_viewer(tv)
        assert len(marks) == 2
        after_update = marks[0].y
        assert not np.allclose(before_update, after_update)

        orig_flux_column = f._obj.flux_column.selected
        assert f.dataset.selected_obj is not None
        assert f._obj.flux_label_overwrite is False
        assert f._obj.flux_label.value == f'{orig_flux_column}_flattened'
        f._obj.vue_apply(add_data=True)
        assert f._obj.flux_column.selected == f'{orig_flux_column}_flattened'
        assert f._obj.flatten_err == ''

    # marks are hidden
    assert len(_get_marks_from_viewer(tv)) == 0
    assert len(_get_marks_from_viewer(pv)) == 0


def test_unnormalize(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)

    f = helper.plugins['Flatten']

    assert f.unnormalize is False
    lc, trend = f.flatten(add_data=False)
    # should be roughly 1
    assert np.nanmedian(lc.flux.value) < 2

    f.unnormalize = True
    lc_unnorm, trend = f.flatten(add_data=False)
    assert_allclose(np.nanmedian(lc.flux.value), np.nanmedian(trend.flux.value), rtol=1e-4)



from glue.core.message import Message

__all__ = ['EphemerisComponentChangedMessage',
           'EphemerisChangedMessage',
           'FluxColumnChangedMessage']


class EphemerisComponentChangedMessage(Message):
    """Message emitted when an ephemeris component is added/renamed/removed in the
    ephemeris plugin"""
    def __init__(self, old_lbl, new_lbl, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.old_lbl = old_lbl
        self.new_lbl = new_lbl
        if old_lbl is not None and new_lbl is not None:
            self.type = 'rename'
        elif old_lbl is None and new_lbl is not None:
            self.type = 'add'
        elif old_lbl is not None and new_lbl is None:
            self.type = 'remove'
        else:  # pragma: no cover
            raise ValueError("must provide at least one of old_lbl or new_lbl")


class EphemerisChangedMessage(Message):
    """Message emitted when the parameters of an ephemeris are updated/changed
    in the ephemeris plugin"""
    def __init__(self, ephemeris_label, *args, **kwargs):
        self.ephemeris_label = ephemeris_label


class FluxColumnChangedMessage(Message):
    """Message emitted by the FluxColumnSelect component when the selection has been changed.
    To subscribe to a change for a particular dataset, consider using FluxColumnSelect directly
    and observing the traitlet, rather than subscribing to this message"""
    def __init__(self, dataset, flux_column, *args, **kwargs):
        self.dataset = dataset
        self.flux_column = flux_column



import pytest


def test_docs_snippets(helper, light_curve_like_kepler_quarter):
    lcviz, lc = helper, light_curve_like_kepler_quarter

    lcviz.load(lc)
    # lcviz.show()

    ephem = lcviz.plugins['Ephemeris']
    ephem.period = 4.88780258
    ephem.t0 = 2.43
    ephem.rename_component('default', 'my component name')


def test_plugin_ephemeris(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)
    ephem = helper.plugins['Ephemeris']

    assert len(helper.app.get_viewer_ids()) == 1
    assert not ephem._obj.phase_viewer_exists

    # setting any value in the ephemeris should create a new viewer
    ephem.period = 3.14
    assert len(helper.app.get_viewer_ids()) == 2
    assert ephem._obj.phase_viewer_exists
    assert 'flux-vs-phase:default' in helper.app.get_viewer_ids()

    ephem.t0 = 5
    ephem._obj.vue_period_double()
    assert ephem.period == 3.14 * 2
    ephem._obj.vue_period_halve()
    assert ephem.period == 3.14

    pv = ephem._obj.default_phase_viewer
    # original limits are set to 0->1 (technically 1-phase_wrap -> phase_wrap)
    assert (pv.state.x_min, pv.state.x_max) == (0.0, 1.0)
    ephem.wrap_at = 0.5
    assert (pv.state.x_min, pv.state.x_max) == (-0.5, 0.5)

    ephem.add_component('custom component')
    assert not ephem._obj.phase_viewer_exists
    ephem.create_phase_viewer()
    assert len(helper.app.get_viewer_ids()) == 3
    assert len(ephem.ephemerides) == 2
    assert 'custom component' in ephem.ephemerides

    with pytest.raises(ValueError):
        # brackets interfere with cloned viewer label logic
        ephem.rename_component('custom component', 'custom component[blah]')
    with pytest.raises(ValueError):
        # colons interfere with viewer ephemeris logic
        ephem.rename_component('custom component', 'custom component:blah')

    ephem.rename_component('custom component', 'renamed custom component')
    assert len(ephem.ephemerides) == 2
    assert 'renamed custom component' in ephem.ephemerides
    assert len(helper.app.get_viewer_ids()) == 3

    assert ephem.component == 'renamed custom component'
    assert ephem.period == 3.14
    assert ephem.ephemeris['period'] == 3.14
    # modify the ephemeris of the NON-selected ephemeris component
    ephem.update_ephemeris(ephem_component='default', period=2)
    assert ephem.period == 3.14
    assert ephem.ephemerides['default']['period'] == 2

    ephem.remove_component('renamed custom component')
    assert len(ephem.ephemerides) == 1
    assert len(helper.app.get_viewer_ids()) == 2
    assert ephem.component == 'default'
    assert ephem.period == 2

    assert ephem.method.selected == 'Lomb-Scargle'
    ephem.method = 'Box Least Squares'
    assert ephem._obj.method_err == ''
    ephem._obj.vue_adopt_period_at_max_power()
    assert ephem.period != 2

    # test that non-zero dpdt does not crash
    ephem.dpdt = 0.005


def test_cloned_phase_viewer(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)
    ephem = helper.plugins['Ephemeris']

    assert len(ephem._obj._get_phase_viewers()) == 0
    pv1 = ephem.create_phase_viewer()
    assert len(ephem._obj._get_phase_viewers()) == 1
    pv2 = pv1._obj.clone_viewer()
    assert len(ephem._obj._get_phase_viewers()) == 2
    assert len(helper.viewers) == 3
    assert pv1._obj.reference_id == 'flux-vs-phase:default'
    assert pv1._obj._ephemeris_component == 'default'
    assert pv2._obj.reference_id == 'flux-vs-phase:default[1]'
    assert pv2._obj._ephemeris_component == 'default'

    # renaming ephemeris should update both labels
    ephem.rename_component('default', 'renamed')
    assert pv1._obj.reference_id == 'flux-vs-phase:renamed'
    assert pv1._obj._ephemeris_component == 'renamed'
    assert pv2._obj.reference_id == 'flux-vs-phase:renamed[1]'
    assert pv2._obj._ephemeris_component == 'renamed'
    assert len(ephem._obj._get_phase_viewers()) == 2

    ephem.remove_component('renamed')
    assert len(helper.viewers) == 1  # just flux-vs-phase


def test_create_phase_viewer(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)
    ephem = helper.plugins['Ephemeris']
    vc = helper._tray_tools['g-viewer-creator']

    assert len(vc.viewer_types) == 2  # time viewer, phase viewer for default
    ephem.adopt_period_at_max_power()
    assert len(vc.viewer_types) == 2  # time viewer, phase viewer at default
    assert len(ephem._obj._get_phase_viewers()) == 1

    vc.vue_create_viewer('flux-vs-phase:default')
    assert len(ephem._obj._get_phase_viewers()) == 2
    for pv in ephem._obj._get_phase_viewers():
        assert pv._ephemeris_component == 'default'

    ephem.rename_component('default', 'renamed')
    assert len(vc.viewer_types) == 2
    vc.vue_create_viewer('flux-vs-phase:renamed')
    assert len(ephem._obj._get_phase_viewers()) == 3

    for pv in ephem._obj._get_phase_viewers():
        assert pv._ephemeris_component == 'renamed'

    ephem.add_component('new')
    assert len(vc.viewer_types) == 3


def test_ephemeris_queries(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)
    ephem = helper.plugins['Ephemeris']

    ephem.query_for_ephemeris()
    planet = ephem.query_result.choices[0]
    assert planet == 'HAT-P-11 b'

    ephem.query_result = planet
    ephem.create_ephemeris_from_query()


def test_ephemeris_query_no_name(helper, light_curve_like_kepler_quarter):
    # test that the query successfully falls back on the RA/Dec:
    light_curve_like_kepler_quarter.meta['OBJECT'] = ''

    helper.load_data(light_curve_like_kepler_quarter)
    ephem = helper.plugins['Ephemeris']

    ephem.query_for_ephemeris()
    planet = ephem.query_result.choices[0]
    assert planet == 'HAT-P-11 b'

    ephem.query_result = planet
    ephem.create_ephemeris_from_query()



from astropy import units as u
from functools import cached_property

from ipyvuetify import VuetifyTemplate
from glue.core import HubListener
from lightkurve import LightCurve
from traitlets import List, Unicode

from jdaviz.core.template_mixin import DatasetSelect, SelectPluginComponent

from lcviz.events import (EphemerisComponentChangedMessage,
                          FluxColumnChangedMessage)

__all__ = ['EphemerisSelect',
           'EphemerisSelectMixin', 'EphemerisSelectAllowNoneMixin',
           'FluxColumnSelect', 'FluxColumnSelectMixin']


class EphemerisSelect(SelectPluginComponent):
    """
    Plugin select for ephemeris components defined by the Ephemeris plugin.

    Useful API methods/attributes:

    * :meth:`~SelectPluginComponent.choices`
    * ``selected``
    * :attr:`selected_obj`
    * :meth:`~SelectPluginComponent.select_default`
    """

    """
    Traitlets (in the object, custom traitlets in the plugin):

    * ``items`` (list of dicts with keys: label)
    * ``selected`` (string)

    Properties (in the object only):

    * ``selected_obj``

    To use in a plugin:

    * create traitlets with default values
    * register with all the automatic logic in the plugin's init by passing the string names
      of the respective traitlets
    * use component in plugin template (see below)
    * refer to properties above based on the interally stored reference to the
      instantiated object of this component

    Example template (label and hint are optional)::

      <plugin-ephemeris-select
        :items="ephemeris_items"
        :selected.sync="ephemeris_selected"
        label="Ephemeris"
        hint="Select ephemeris."
      />

    """
    def __init__(self, plugin, items, selected,
                 default_text='No ephemeris', manual_options=[],
                 default_mode='first'):
        """
        Parameters
        ----------
        plugin
            the parent plugin object
        items : str
            the name of the items traitlet defined in ``plugin``
        selected : str
            the name of the selected traitlet defined in ``plugin``
        default_text : str or None
            the text to show for no selection.  If None, no entry will be provided
            in the dropdown for no selection.
        manual_options: list
            list of options to provide that are not automatically populated by ephemerides.  If
            ``default`` text is provided but not in ``manual_options`` it will still be included as
            the first item in the list.
        """
        super().__init__(plugin, items=items, selected=selected,
                         default_text=default_text, manual_options=manual_options,
                         default_mode=default_mode)
        self.hub.subscribe(self, EphemerisComponentChangedMessage,
                           handler=self._ephem_component_change)

    @cached_property
    def ephemeris_plugin(self):
        return self.app._jdaviz_helper.plugins.get('Ephemeris', None)

    @cached_property
    def selected_obj(self):
        if self.selected in self._manual_options:
            return None
        return self.ephemeris_plugin.ephemerides.get(self.selected, None)

    def get_data_for_dataset(self, dataset, ycomp='flux'):
        if not isinstance(dataset, DatasetSelect):  # pragma: no cover
            raise ValueError("dataset must be DatasetSelect object")
        if self.selected in self._manual_options:
            return dataset.selected_obj
        return self.ephemeris_plugin.get_data(dataset.selected, self.selected)

    def _ephem_component_change(self, msg=None):
        type = getattr(msg, 'type', None)
        if type == 'remove' and msg.old_lbl in self.choices:
            self.items = [item for item in self.items if item['label'] != msg.old_lbl]
            self._apply_default_selection()
        elif type == 'rename' and msg.old_lbl in self.choices:
            was_selected = self.selected == msg.old_lbl
            self.items = [item if item['label'] != msg.old_lbl else {'label': msg.new_lbl}
                          for item in self.items]
            if was_selected:
                self.selected = msg.new_lbl
        elif type == 'add' and msg.new_lbl not in self.choices:
            self.items = self.items + [{'label': msg.new_lbl}]
        else:
            # something might be out of sync, build from scratch
            manual_items = [{'label': label} for label in self.manual_options]
            self.items = manual_items + [{'label': component}
                                         for component in self.ephemeris_plugin.ephemerides.keys()]
            self._apply_default_selection()


class EphemerisSelectMixin(VuetifyTemplate, HubListener):
    """
    Applies the EphemerisSelect component as a mixin in the base plugin.  This
    automatically adds traitlets as well as new properties to the plugin with minimal
    extra code.  For multiple instances or custom traitlet names/defaults, use the
    component instead.

    To use in a plugin:

    * add ``EphemerisSelectMixin`` as a mixin to the class
    * use the traitlets available from the plugin or properties/methods available from
      ``plugin.ephemeris``.

    Example template (label and hint are optional)::

      <plugin-ephemeris-select
        :items="ephemeris_items"
        :selected.sync="ephemeris_selected"
        label="Ephemeris"
        hint="Select ephemeris."
      />

    """
    ephemeris_items = List().tag(sync=True)
    ephemeris_selected = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ephemeris = EphemerisSelect(self, 'ephemeris_items', 'ephemeris_selected',
                                         default_text=None)


class EphemerisSelectAllowNoneMixin(VuetifyTemplate, HubListener):
    """
    Applies the EphemerisSelect component as a mixin in the base plugin.  This
    automatically adds traitlets as well as new properties to the plugin with minimal
    extra code.  For multiple instances or custom traitlet names/defaults, use the
    component instead.

    To use in a plugin:

    * add ``EphemerisSelectMixin`` as a mixin to the class
    * use the traitlets available from the plugin or properties/methods available from
      ``plugin.ephemeris``.

    Example template (label and hint are optional)::

      <plugin-ephemeris-select
        :items="ephemeris_items"
        :selected.sync="ephemeris_selected"
        label="Ephemeris"
        hint="Select ephemeris."
      />

    """
    ephemeris_items = List().tag(sync=True)
    ephemeris_selected = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ephemeris = EphemerisSelect(self, 'ephemeris_items', 'ephemeris_selected')


class FluxColumnSelect(SelectPluginComponent):
    def __init__(self, plugin, items, selected, dataset):
        super().__init__(plugin,
                         items=items,
                         selected=selected,
                         dataset=dataset)

        self.dataset.get_data_cls = LightCurve

        self.add_observe(selected, self._on_change_selected)
        self.add_observe(self.dataset._plugin_traitlets['selected'],
                         self._on_change_dataset)

        # sync between instances in different plugins
        self.hub.subscribe(self, FluxColumnChangedMessage,
                           handler=self._on_flux_column_changed_msg)

    def _on_change_dataset(self, *args):
        def _include_col(lk_obj, col):
            if col == 'flux' and lk_obj.meta.get('FLUX_ORIGIN') != 'flux':
                # this is the currently active column (and should be copied elsewhere unless)
                return False
            if col in ('time', 'cadn', 'cadenceno', 'quality'):
                return False
            if col.startswith('phase:'):
                # internal jdaviz ephemeris phase columns
                return False
            if col.startswith('time'):
                return False
            if col.startswith('centroid'):
                return False
            if col.startswith('cbv'):
                # cotrending basis vector
                return False
            if col.endswith('_err'):
                return False
            if col.endswith('quality'):
                return False
            # TODO: need to think about flatten losing units in the flux column
            return lk_obj[col].unit != u.pix

        lk_obj = self.dataset.selected_obj
        if lk_obj is None:
            return
        self.choices = [col for col in lk_obj.columns if _include_col(lk_obj, col)]
        flux_column = lk_obj.meta.get('FLUX_ORIGIN')
        if flux_column in self.choices:
            self.selected = flux_column
        else:
            self.selected = ''

    def _on_flux_column_changed_msg(self, msg):
        if msg.dataset != self.dataset.selected:
            return

        # need to clear the cache due to the change in metadata made to the data-collection entry
        self.dataset._clear_cache('selected_obj', 'selected_dc_item')
        self._on_change_dataset()
        self.selected = msg.flux_column

    def _on_change_selected(self, *args):
        if self.selected == '':
            return

        dc_item = self.dataset.selected_dc_item
        old_flux_column = dc_item.meta.get('FLUX_ORIGIN')
        if self.selected == old_flux_column:
            # nothing to do here!
            return

        # instead of using lightkurve's select_flux and having to reparse the data entry, we'll
        # manipulate the arrays in the data-collection directly, and modify FLUX_ORIGIN so that
        # exporting back to a lightkurve object works as expected
        self.app._jdaviz_helper._set_data_component(dc_item, 'flux', dc_item[self.selected])
        if self.selected+"_err" in dc_item.component_ids():
            if "flux_err" in dc_item.component_ids():
                self.app._jdaviz_helper._set_data_component(dc_item, 'flux_err',
                                                            dc_item[self.selected + "_err"])
            else:
                dc_item.add_component(dc_item[self.selected + "_err"], 'flux_err')
        else:
            dc_item.remove_component(dc_item.find_component_id('flux_err'))

        dc_item.meta['FLUX_ORIGIN'] = self.selected

        self.hub.broadcast(FluxColumnChangedMessage(dataset=self.dataset.selected,
                                                    flux_column=self.selected,
                                                    sender=self))

    def add_new_flux_column(self, flux, flux_err, label, selected=False):
        dc_item = self.dataset.selected_dc_item
        self.app._jdaviz_helper._set_data_component(dc_item,
                                                    label,
                                                    flux)
        self.app._jdaviz_helper._set_data_component(dc_item,
                                                    f"{label}_err",
                                                    flux_err)

        # broadcast so all instances update to get the new column and selection (if applicable)
        self.hub.broadcast(FluxColumnChangedMessage(dataset=self.dataset.selected,
                                                    flux_column=label if selected else self.selected,  # noqa
                                                    sender=self))


class FluxColumnSelectMixin(VuetifyTemplate, HubListener):
    flux_column_items = List().tag(sync=True)
    flux_column_selected = Unicode().tag(sync=True)
    # assumes DatasetSelectMixin is also used (DatasetSelectMixin must appear after in inheritance)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.flux_column = FluxColumnSelect(self,
                                            'flux_column_items',
                                            'flux_column_selected',
                                            dataset='dataset')



import numpy as np

from traitlets import observe
from jdaviz.configs.default.plugins import PlotOptions
from jdaviz.core.registries import tray_registry

__all__ = ['PlotOptions']


@tray_registry('plot-options', label="Plot Options",
               category='core', sidebar='viewers', subtab=0, overwrite=True)
class PlotOptions(PlotOptions):
    """
    See the :ref:`Plot Options Plugin Documentation <plot-options>` for more details.

    For a full list of exposed attributes, call ``dir(plugin)``.  Note that some attributes are
    applicable depending on the selection of ``viewer`` and/or ``layer``.  Below are
    a list of some common attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``viewer`` (:class:`~jdaviz.core.template_mixin.ViewerSelect`):
    * ``viewer_multiselect``
    * ``layer`` (:class:`~jdaviz.core.template_mixin.LayerSelect`):
    * ``layer_multiselect``
    * :meth:`select_all`
    * ``subset_color`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
    * ``line_color`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
    * ``line_width`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
    * ``line_opacity`` (:class:`~jdaviz.core.template_mixin.PlotOptionsSyncState`):
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @observe('vdocs')
    def _update_docs_link(self, *args):
        self.docs_link = f"https://lcviz.readthedocs.io/en/{self.vdocs}/plugins.html#plot-options"

    def _default_tpf_stretch(
            self, vmin_percentile=5, vmax_percentile=99, tpf_viewer_reference='image'
    ):
        viewer = self.app.get_viewer(tpf_viewer_reference)
        image = viewer.layers[0].get_image_data()
        vmin, vmax = np.nanpercentile(
            image, [vmin_percentile, vmax_percentile]
        )

        self.viewer_selected = tpf_viewer_reference
        self.stretch_function_value = 'log'
        self.stretch_vmin_value = vmin
        self.stretch_vmax_value = vmax

    @property
    def user_api(self):
        api = super().user_api
        expose = [e for e in api._expose if e not in ('apply_RGB_presets', 'line_as_steps',
                  'uncertainty_visible')]

        expose += ['marker_visible', 'marker_fill', 'marker_opacity',
                   'marker_size_mode', 'marker_size', 'marker_size_scale',
                   'marker_size_col', 'marker_size_vmin', 'marker_size_vmax',
                   'marker_color_mode', 'marker_color', 'marker_color_col',
                   'marker_colormap', 'marker_colormap_vmin', 'marker_colormap_vmax',
                   'line_visible', 'line_width']
        api._expose = expose

        return api



from lightkurve import KeplerTargetPixelFile, TessTargetPixelFile

from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.loaders.importers import BaseImporterToDataCollection

from lcviz.viewers import TimeScatterView, PhaseScatterView


__all__ = ['TPFImporter']


@loader_importer_registry('TPF')
class TPFImporter(BaseImporterToDataCollection):
    template_file = __file__, "../to_dc_with_label.vue"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.is_valid:
            self.data_label_default = f"{self.input.meta.get('OBJECT', 'Light curve')} [TPF]"

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'lcviz'):
            # NOTE: temporary during deconfig process
            return False
        return isinstance(self.input, (KeplerTargetPixelFile, TessTargetPixelFile))

    @staticmethod
    def _get_supported_viewers():
        return [{'label': 'image', 'reference': 'lcviz-cube-viewer'}]

    @property
    def ignore_viewers_with_cls(self):
        # time viewers do support manually loading TPF data, but we do not want to
        # load automatically and instead default to creating and loading into
        # an image viewer
        return (TimeScatterView, PhaseScatterView)

    @property
    def output(self):
        return self.input



def test_docs_snippets(helper, light_curve_like_kepler_quarter):
    lcviz, lc = helper, light_curve_like_kepler_quarter

    lcviz.load(lc)
    # lcviz.show()

    po = lcviz.plugins['Plot Options']
    print(f"viewer choices: {po.viewer.choices}")
    po.viewer = po.viewer.choices[0]
    print(f"layer choices: {po.layer.choices}")
    po.layer = po.layer.choices[0]

    po.marker_size = 4
    po.marker_color = 'blue'



import numpy as np

from glue.core.subset_group import GroupedSubset
from jdaviz.configs.imviz.plugins.coords_info import CoordsInfo
from jdaviz.core.events import ViewerRenamedMessage
from jdaviz.core.registries import tool_registry

from lcviz.viewers import TimeScatterView, PhaseScatterView, CubeView

__all__ = ['CoordsInfo']


@tool_registry('g-coords-info', overwrite=True)
class CoordsInfo(CoordsInfo):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._supported_viewer_classes = self._supported_viewer_classes + (
            TimeScatterView, PhaseScatterView, CubeView
        )
        self._viewer_classes_with_marker = self._viewer_classes_with_marker + (
            TimeScatterView, PhaseScatterView
        )

        # TODO: move to jdaviz if/once viewer renaming supported
        self.hub.subscribe(self, ViewerRenamedMessage,
                           handler=self._viewer_renamed)

    def _viewer_renamed(self, msg):
        self._marks[msg.new_viewer_ref] = self._marks.pop(msg.old_viewer_ref)

    def _image_shape_inds(self, image):
        if image.ndim == 3:
            # exception to the upstream cubeviz case of (0, 1)
            return (2, 1)
        return super()._image_shape_inds(image)

    def _get_cube_value(self, image, arr, x, y, viewer):
        if image.ndim == 3:
            # exception to the upstream cubeviz case of x, y, slice
            return arr[viewer.state.slices[0], int(round(y)), int(round(x))]
        return super()._get_cube_value(image, arr, x, y, viewer)

    def _lc_viewer_update(self, viewer, x, y, mouseevent=True):
        is_phase = isinstance(viewer, PhaseScatterView)
        # TODO: update with display_unit when supported in lcviz
        x_unit = '' if is_phase else str(viewer.time_unit)
        y_unit = str(viewer.data()[0].flux.unit)

        def _cursor_fallback():
            self._dict['axes_x'] = x
            self._dict['axes_x:unit'] = x_unit
            self._dict['axes_y'] = y
            self._dict['axes_y:unit'] = y_unit

            self._dict['data_label'] = ''
            self._dict['time'] = x if not is_phase else np.nan
            self._dict['time:unit'] = x_unit
            self._dict['phase'] = x if is_phase else np.nan
            self._dict['value'] = y
            self._dict['value:unit'] = y_unit
            self._dict['ephemeris'] = ''

            self.row2_title = ''
            self.row2_text = ''
            self.row3_title = ''
            self.row3_text = ''
            self.icon = 'mdi-cursor-default'
            self.marks[viewer._reference_id].visible = False

        self.row1a_title = 'Cursor'
        self.row1a_text = f'{x:10.5e}, {y:10.5e}'

        # show the locked marker/coords only if either no tool or the default tool is active
        if self.dataset.selected == 'none':
            _cursor_fallback()
            return

        xrange = abs(viewer.state.x_max - viewer.state.x_min)
        yrange = abs(viewer.state.y_max - viewer.state.y_min)

        # Snap to the closest data point
        closest_distsq = None
        closest_x = None
        closest_y = None
        closest_icon = None
        closest_lyr = None
        for lyr in viewer.layers:
            if isinstance(lyr.layer, GroupedSubset):
                continue
            if self.dataset.selected == 'auto' and not lyr.visible:
                continue
            if self.dataset.selected != 'auto' and self.dataset.selected != lyr.layer.label:
                continue

            scatter = lyr.scatter_mark
            lyr_x, lyr_y = scatter.x, scatter.y
            if not len(lyr_x):
                continue

            # NOTE: unlike specviz which determines the closest point in x per-layer,
            # this determines the closest point in x/y per-layer in pixel-space
            # (making it easier to get the snapping point into shallow eclipses, etc)
            distsqs = ((lyr_x - x)/xrange)**2 + ((lyr_y - y)/yrange)**2
            cur_i = np.nanargmin(distsqs)
            cur_x, cur_y = float(lyr_x[cur_i]), float(lyr_y[cur_i])
            cur_distsq = distsqs[cur_i]

            if (closest_distsq is None) or (cur_distsq < closest_distsq):
                closest_distsq = cur_distsq
                closest_i = cur_i
                closest_x = cur_x
                closest_y = cur_y
                closest_icon = self.app.state.layer_icons.get(lyr.layer.label, '')
                closest_lyr = lyr
                self._dict['data_label'] = lyr.layer.label

        if closest_lyr is None:
            _cursor_fallback()
            return

        self.row2_title = 'Phase' if is_phase else 'Time'
        if is_phase:
            self.row2_text = f'{closest_x:0.05f}'
            component_labels = [comp.label for comp in closest_lyr.layer.components]
            time_comp = closest_lyr.layer.components[component_labels.index('World 0')]
            times = closest_lyr.layer.get_data(time_comp)
            self._dict['time'] = float(times[closest_i])
            self._dict['time:unit'] = str(viewer.time_unit)  # x_unit is phase
            self._dict['phase'] = closest_x
            self._dict['ephemeris'] = viewer.reference.split(':')[1]
        else:
            self.row2_text = f'{closest_x:10.5e} {x_unit}'
            self._dict['time'] = closest_x
            self._dict['time:unit'] = x_unit
            self._dict['phase'] = np.nan
            self._dict['ephemeris'] = ''

        self._dict['axes_x'] = closest_x
        self._dict['axes_x:unit'] = x_unit
        self._dict['index'] = float(closest_i)

        self.row3_title = 'Flux'
        self.row3_text = f'{closest_y:10.5e} {y_unit}'
        self._dict['axes_y'] = closest_y
        self._dict['axes_y:unit'] = y_unit
        self._dict['value'] = closest_y
        self._dict['value:unit'] = y_unit

        self.icon = closest_icon

        self.marks[viewer._reference_id].update_xy([closest_x], [closest_y])  # noqa
        self.marks[viewer._reference_id].visible = True

    def _image_viewer_update(self, viewer, x, y, mouseevent=True):
        # Extract first dataset from visible layers and use this for coordinates - the choice
        # of dataset shouldn't matter if the datasets are linked correctly
        active_layer = viewer.active_image_layer
        if active_layer is None:
            self._viewer_mouse_clear_event(viewer)
            return

        # TODO: refactor this code block (to retrieve the active layer)
        # upstream to make it resuable
        from glue_jupyter.bqplot.image.layer_artist import BqplotImageSubsetLayerArtist

        if self.dataset.selected == 'auto':
            image = active_layer.layer
        elif self.dataset.selected == 'none':
            active_layer = viewer.layers[0].state
            image = viewer.layers[0].layer
        else:
            for layer in viewer.layers:
                if layer.layer.label == self.dataset.selected and layer.visible:
                    if isinstance(layer, BqplotImageSubsetLayerArtist):
                        # cannot expose info for spatial subset layers
                        continue
                    active_layer = layer.state
                    image = layer.layer
                    break
            else:
                image = None

        self._dict['axes_x'] = x
        self._dict['axes_x:unit'] = 'pix'
        self._dict['axes_y'] = y
        self._dict['axes_y:unit'] = 'pix'

        # set default empty values
        if self.dataset.selected != 'none' and image is not None:
            self.icon = self.app.state.layer_icons.get(image.label, '')  # noqa
            self._dict['data_label'] = image.label

        if self.dataset.selected == 'none' or image is None:
            self.icon = 'mdi-cursor-default'
            self._dict['data_label'] = ''
        else:
            try:
                time = viewer.slice_value
            except IndexError:
                self._viewer_mouse_clear_event(viewer)
                return
            if time is None:
                self._viewer_mouse_clear_event(viewer)
                return
            # TODO: store slice unit within image viewer to avoid this assumption?
            tvs = self.app.get_viewers_of_cls(TimeScatterView)
            if len(tvs):
                time_unit = str(tvs[0].time_unit)
            else:
                time_unit = 's'
            self.row2_title = 'Time'
            self.row2_text = f'{time:0.5f} {time_unit}'
            self._dict['time'] = time
            self._dict['time:unit'] = time_unit

        maxsize = int(np.ceil(np.log10(np.max(active_layer.layer.shape)))) + 3
        fmt = 'x={0:0' + str(maxsize) + '.1f} y={1:0' + str(maxsize) + '.1f}'
        self.row1a_title = 'Pixel'
        self.row1a_text = (fmt.format(x, y))
        self._dict['pixel'] = (float(x), float(y))

        if self.dataset.selected == 'none' or image is None:
            # no data values to extract
            self.row1b_title = ''
            self.row1b_text = ''
            return

        # Extract data values at this position.
        # Check if shape is [x, y, z] or [y, x] and show value accordingly.
        ix_shape, iy_shape = self._image_shape_inds(image)

        if (-0.5 < x < image.shape[ix_shape] - 0.5 and -0.5 < y < image.shape[iy_shape] - 0.5
                and hasattr(active_layer, 'attribute')):
            attribute = active_layer.attribute
            arr = image.get_component(attribute).data
            unit = image.get_component(attribute).units
            value = self._get_cube_value(image, arr, x, y, viewer)
            self.row1b_title = 'Value'
            self.row1b_text = f'{value:+10.5e} {unit}'
            self._dict['value'] = float(value)
            self._dict['value:unit'] = unit
        else:
            self.row1b_title = ''
            self.row1b_text = ''

    def update_display(self, viewer, x, y, mouseevent=True):
        self._dict = {}
        self.app.state.show_toolbar_buttons = False
        if not len(viewer.state.layers):
            return

        if isinstance(viewer, (TimeScatterView, PhaseScatterView)):
            self._lc_viewer_update(viewer, x, y, mouseevent)
        elif isinstance(viewer, CubeView):
            self._image_viewer_update(viewer, x, y, mouseevent)



from traitlets import observe

from jdaviz.configs.default.plugins import Export
from jdaviz.core.registries import tray_registry

__all__ = ['Export']


@tray_registry('export', label="Export",
               category='core', sidebar='save', overwrite=True)
class Export(Export):
    """
    See the :ref:`Export Plot Plugin Documentation <export>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``viewer`` (:class:`~jdaviz.core.template_mixin.ViewerSelect`)
    * ``viewer_format`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`)
    * ``filename``
    * :meth:`export`
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @observe('vdocs')
    def _update_docs_link(self, *args):
        self.docs_link = f"https://lcviz.readthedocs.io/en/{self.vdocs}/plugins.html#export"



from glue.core.message import (DataCollectionAddMessage,
                               DataCollectionDeleteMessage)
from jdaviz.configs.default.plugins import ViewerCreator
from jdaviz.core.events import NewViewerMessage
from jdaviz.core.registries import tool_registry
from lcviz.events import EphemerisComponentChangedMessage
from lcviz.viewers import TimeScatterView, CubeView

__all__ = ['ViewerCreator']


# overwrite requires upstream changes, we can do without if we just lose the tooltip
@tool_registry('g-viewer-creator', overwrite=True)
class ViewerCreator(ViewerCreator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for msg in (EphemerisComponentChangedMessage,
                    DataCollectionAddMessage,
                    DataCollectionDeleteMessage):
            self.hub.subscribe(self, msg,
                               handler=lambda x: self._rebuild_available_viewers())
        self._rebuild_available_viewers()

    def _rebuild_available_viewers(self, *args):
        # filter to lcviz-specific viewers only
        # list of dictionaries with name (registry name)
        # and label (what appears in dropdown and the default label of the viewer)

        if self.app._jdaviz_helper is not None:
            try:
                # NOTE: if first data was just added, the plugin may still be irrelevant
                # and so not available from .plugins dictionary
                ephem_plg = self.app.get_tray_item_from_name('ephemeris')
            except KeyError:
                ephem_plg = None
            if ephem_plg is not None:
                phase_viewers = [{'name': f'lcviz-phase-viewer:{e}', 'label': f'flux-vs-phase:{e}'}
                                for e in ephem_plg.component.choices]  # noqa
            else:
                phase_viewers = []
            if self.app._jdaviz_helper._has_cube_data:
                cube_viewers = [{'name': 'lcviz-cube-viewer', 'label': 'image'}]
            else:
                cube_viewers = []
        else:
            phase_viewers = [{'name': 'lcviz-phase-viewer:default',
                              'label': 'flux-vs-phase:default'}]
            cube_viewers = []

        self.viewer_types = [v for v in self.viewer_types if v['name'].startswith('lcviz')
                             and not v['label'].startswith('flux-vs-phase')
                             and not v['label'] in ('cube', 'image')] + phase_viewers + cube_viewers
        self.send_state('viewer_types')

    def vue_create_viewer(self, name):
        if name.startswith('lcviz-phase-viewer') or name.startswith('flux-vs-phase'):
            ephem_comp = name.split(':')[1]
            ephem_plg = self.app._jdaviz_helper.plugins['Ephemeris']
            ephem_plg.create_phase_viewer(ephem_comp)
            return
        if name in ('flux-vs-time', 'lcviz-time-viewer'):
            # allow passing label and map to the name for upstream support
            viewer_id = self.app._jdaviz_helper._get_clone_viewer_reference('flux-vs-time')
            self.app._on_new_viewer(NewViewerMessage(TimeScatterView, data=None, sender=self.app),
                                    vid=viewer_id, name=viewer_id)
            return
        if name in ('image', 'lcviz-cube-viewer'):
            viewer_id = self.app._jdaviz_helper._get_clone_viewer_reference('image')
            self.app._on_new_viewer(NewViewerMessage(CubeView, data=None, sender=self.app),
                                    vid=viewer_id, name=viewer_id)
            return

        super().vue_create_viewer(name)



from traitlets import observe

from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        DatasetSelectMixin)
from jdaviz.core.user_api import PluginUserApi

from lcviz.components import FluxColumnSelectMixin
from lcviz.utils import is_lc

__all__ = ['FluxColumn']


@tray_registry('flux-column', label="Flux Column", category='app:options')
class FluxColumn(PluginTemplateMixin, FluxColumnSelectMixin, DatasetSelectMixin):
    """
    See the :ref:`Flux Column Plugin Documentation <flux-column>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Dataset to bin.
    * ``flux_column`` (:class:`~lcviz.components.FluxColumnSelect`)
    """
    template_file = __file__, "flux_column.vue"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Choose flux column.'

        # NOTE: may eventually want to add support for choosing the column for TPFs
        self.dataset.add_filter(is_lc)

        self._set_relevant()

    @observe('dataset_items')
    def _set_relevant(self, *args):
        if not len(self.dataset_items):
            self.irrelevant_msg = 'No valid datasets loaded'
        else:
            self.irrelevant_msg = ''

    @property
    def user_api(self):
        expose = ['dataset', 'flux_column']
        return PluginUserApi(self, expose=expose)



from .viewer_creator.viewer_creator import *  # noqa
from .coords_info.coords_info import *  # noqa

from .binning.binning import *  # noqa
from .ephemeris.ephemeris import *  # noqa
from .export.export import *  # noqa
from .flatten.flatten import *  # noqa
from .flux_column.flux_column import *  # noqa
from .frequency_analysis.frequency_analysis import *  # noqa
from .markers.markers import *  # noqa
from .time_selector.time_selector import *  # noqa
from .metadata_viewer.metadata_viewer import *  # noqa
from .plot_options.plot_options import *  # noqa
from .stitch.stitch import *  # noqa
from .subset_tools.subset_tools import *  # noqa



import pytest

from lcviz.marks import LivePreviewBinning


def _get_marks_from_viewer(viewer, cls=(LivePreviewBinning)):
    return [m for m in viewer.figure.marks if isinstance(m, cls) and m.visible]


def test_docs_snippets(helper, light_curve_like_kepler_quarter):
    lcviz, lc = helper, light_curve_like_kepler_quarter

    lcviz.load(lc)
    # lcviz.show()

    binning = lcviz.plugins['Binning']
    binning.n_bins = 150
    binned_lc = binning.bin(add_data=True)
    print(binned_lc)


def test_plugin_binning(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)
    tv = helper.default_time_viewer._obj

    b = helper.plugins['Binning']
    b._obj.plugin_opened = True
    ephem = helper.plugins['Ephemeris']
    ephem.period = 1.2345
    # pv = ephem.create_phase_viewer()._obj

    with b.as_active():
        b.show_live_preview = True
        assert b.ephemeris == 'No ephemeris'
        assert len(_get_marks_from_viewer(tv)) == 1
        # NOTE: this seems to be failing on some machines but not others; skipping for now
        # assert len(_get_marks_from_viewer(pv)) == 1
        assert b._obj.ephemeris_dict == {}

        # update ephemeris will force re-phasing
        ephem.period = 1.111

        b.bin(add_data=True)

        b.ephemeris = 'default'
        # NOTE: this seems to be failing on some machines but not others; skipping for now
        # assert len(_get_marks_from_viewer(tv)) == 0
        # assert len(_get_marks_from_viewer(pv)) == 1
        assert len(b._obj.ephemeris_dict.keys()) > 0

        # update ephemeris will force re-binning
        ephem.period = 1.222

        b.bin(add_data=True)

        # setting to invalid n_bins will raise error
        b.n_bins = 0
        # NOTE: this seems to be failing on some machines but not others; skipping for now
        # assert b._obj.bin_enabled is False
        with pytest.raises(ValueError):
            b.bin(add_data=False)

        # the enabled state of the button should work with or without live previews enabled
        b.show_live_preview = False
        assert len(_get_marks_from_viewer(tv)) == 0
        # NOTE: this seems to be failing on some machines but not others; skipping for now
        # assert len(_get_marks_from_viewer(pv)) == 0
        # assert b._obj.bin_enabled is False
        # b.n_bins = 1
        # assert b._obj.bin_enabled is True
        # b.n_bins = ''
        # assert b._obj.bin_enabled is False



import numpy as np
from functools import cached_property
from traitlets import Bool, Float, List, Unicode, observe

from lightkurve import periodogram

from jdaviz.core.custom_traitlets import FloatHandleEmpty
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        DatasetSelectMixin, SelectPluginComponent, PlotMixin)
from jdaviz.core.user_api import PluginUserApi

from lcviz.utils import data_not_folded, is_lc


__all__ = ['FrequencyAnalysis']


@tray_registry('frequency-analysis', label="Frequency Analysis", category='data:analysis')
class FrequencyAnalysis(PluginTemplateMixin, DatasetSelectMixin, PlotMixin):
    """
    See the :ref:`Frequency Analysis Plugin Documentation <frequency_analysis>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Dataset to use for analysis.
    * ``method`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      Method/algorithm to determine the period.
    * ``xunit`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      Whether to plot power vs fequency or period.
    * ``auto_range`` : bool
    * ``minimum`` : float
    * ``maximum`` : float
    * :meth:``periodogram``
    """
    template_file = __file__, "frequency_analysis.vue"

    method_items = List().tag(sync=True)
    method_selected = Unicode().tag(sync=True)

    xunit_items = List().tag(sync=True)
    xunit_selected = Unicode().tag(sync=True)

    auto_range = Bool(True).tag(sync=True)
    minimum = FloatHandleEmpty(0.1).tag(sync=True)  # frequency
    minimum_step = Float(0.1).tag(sync=True)
    maximum = FloatHandleEmpty(1).tag(sync=True)  # frequency
    maximum_step = Float(0.1).tag(sync=True)

    spinner = Bool().tag(sync=True)
    err = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._plugin_description = 'Frequency/period analysis.'
        self._ignore_auto_update = False

        # do not support data only in phase-space
        self.dataset.add_filter(data_not_folded, is_lc)

        self.method = SelectPluginComponent(self,
                                            items='method_items',
                                            selected='method_selected',
                                            manual_options=['Lomb-Scargle', 'Box Least Squares'])

        self.xunit = SelectPluginComponent(self,
                                           items='xunit_items',
                                           selected='xunit_selected',
                                           manual_options=['frequency', 'period'])

        self.plot.figure.axes[1].label = 'power'
        self.plot.figure.fig_margin = {'top': 60, 'bottom': 60, 'left': 65, 'right': 15}
        self.plot.viewer.axis_y.num_ticks = 5
        self.plot.viewer.axis_y.tick_format = '0.2e'
        self.plot.viewer.axis_y.label_offset = '55px'
        self._update_xunit()

        self._set_relevant()

    @observe('dataset_items')
    def _set_relevant(self, *args):
        if not len(self.dataset_items):
            self.irrelevant_msg = 'No valid datasets loaded'
        else:
            self.irrelevant_msg = ''

    @property
    def user_api(self):
        expose = ['dataset', 'method', 'xunit', 'auto_range', 'minimum', 'maximum', 'periodogram']
        return PluginUserApi(self, expose=expose)

    @cached_property
    def periodogram(self):
        # TODO: support multiselect on self.dataset and combine light curves (or would that be a
        # dedicated plugin of its own)?
        self.spinner = True
        self.err = ''
        if self.auto_range:
            min_period, max_period = None, None
        elif self.xunit_selected == 'period':
            min_period, max_period = self.minimum, self.maximum
        else:
            min_period, max_period = self.maximum ** -1, self.minimum ** -1
        if self.method == 'Box Least Squares':
            try:
                per = periodogram.BoxLeastSquaresPeriodogram.from_lightcurve(self.dataset.selected_obj,  # noqa
                                                                             minimum_period=min_period,  # noqa
                                                                             maximum_period=max_period)  # noqa
            except Exception as err:
                self.spinner = False
                self.err = str(err)
                self.plot.update_style('periodogram', visible=False)
                return None
        elif self.method == 'Lomb-Scargle':
            try:
                per = periodogram.LombScarglePeriodogram.from_lightcurve(self.dataset.selected_obj,
                                                                         minimum_period=min_period,
                                                                         maximum_period=max_period)
            except Exception as err:
                self.spinner = False
                self.err = str(err)
                self.plot.update_style('periodogram', visible=False)
                return None
        else:
            self.spinner = False
            raise NotImplementedError(f"periodogram not implemented for {self.method}")

        self._update_periodogram_labels(per)
        self.spinner = False
        return per

    @observe('xunit_selected')
    def _update_xunit(self, *args):
        per = self.periodogram
        if per is not None:
            x = getattr(per, self.xunit_selected).value
            self.plot._update_data('periodogram', x=x)
            self.plot.update_style('periodogram', visible=True)
            old_xmin, old_xmax = self.plot.viewer.state.x_min, self.plot.viewer.state.x_max
            new_xmin = old_xmax ** -1 if old_xmax > 0 else np.nanmin(x)
            new_xmax = old_xmin ** -1 if old_xmin > 0 else np.nanmax(x)
            self.plot.set_limits(x_min=new_xmin, x_max=new_xmax)
        else:
            self.plot.update_style('periodogram', visible=False)

        self._update_periodogram_labels(per)

        # convert minimum/maximum for next computation
        self._ignore_auto_update = True
        old_min, old_max = self.minimum, self.maximum
        self.minimum = old_max ** -1 if old_max > 0 else 0
        self.maximum = old_min ** -1 if old_min > 0 else 0
        self._ignore_auto_update = False

    def _update_periodogram_labels(self, per=None):
        per = per if per is not None else self.periodogram
        if per is not None:
            self.plot.figure.axes[0].label = f"{self.xunit_selected} ({getattr(per, self.xunit_selected).unit})"  # noqa
            self.plot.figure.axes[1].label = f"power ({per.power.unit})" if per.power.unit != "" else "power"  # noqa
        else:
            self.plot.figure.axes[0].label = self.xunit_selected
            self.plot.figure.axes[1].label = "power"

    @observe('dataset_selected', 'method_selected', 'auto_range', 'minimum', 'maximum')
    def _update_periodogram(self, *args):
        if not (hasattr(self, 'method') and hasattr(self, 'dataset')):
            return
        if self._ignore_auto_update:
            return

        # TODO: avoid clearing cache if change was to min/max but auto_range is True?
        self._clear_cache('periodogram')

        per = self.periodogram
        if per is not None:
            self.plot._update_data('periodogram',
                                   x=getattr(per, self.xunit_selected),
                                   y=per.power.value)
            self.plot.update_style(
                'periodogram',
                density_map=False,
                line_visible=True,
                markers_visible=False
            )
            self._update_periodogram_labels(per)
        else:
            self.plot.update_style('periodogram', visible=False)



import os
import numpy as np
from functools import cached_property
from traitlets import Any, Bool, List, Unicode, observe
from astropy.io import fits
from astropy.table import Table
from lightkurve import LightCurve

from glue_jupyter.common.toolbar_vuetify import read_icon

from jdaviz.core.tools import ICON_DIR
from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.loaders.importers import BaseImporterToDataCollection
from jdaviz.core.template_mixin import SelectFileExtensionComponent
from jdaviz.core.user_api import ImporterUserApi
from lcviz.utils import _data_with_reftime


__all__ = ['LightCurveImporter']


def hdu_is_valid(hdu):
    """
    Check if the HDU is a valid light curve HDU.

    Parameters
    ----------
    hdu : `astropy.io.fits.hdu.base.HDUBase`
        The HDU to check.

    Returns
    -------
    bool
        True if the HDU is a valid light curve HDU, False otherwise.
    """
    return (isinstance(hdu, fits.hdu.table.BinTableHDU) and
            'TIME' in hdu.columns.names and
            'LC_INIT' in hdu.columns.names and
            'LC_INIT_ERR' in hdu.columns.names and
            'TUNIT1' in hdu.header)


def hdulist_to_lightcurve(pri_header, hdu):
    data = Table(hdu.data)
    # don't load some columns with names that may
    # conflict with components generated later by lcviz
    for col in ('PHASE', 'CADENCENO'):
        if col in data.columns:
            data.remove_column(col)
    # Remove rows that have NaN data
    data = data[~np.isnan(data['LC_INIT'])]
    time_offset = int(hdu.header['TUNIT1'].split('- ')[1].split(',')[0])
    data['TIME'] += time_offset
    lc = LightCurve(data=data,
                    time=data['TIME'],
                    flux=data['LC_INIT'],
                    flux_err=data['LC_INIT_ERR'])
    lc.meta = dict(pri_header)
    lc.meta = lc.meta | dict(hdu.header)
    lc.meta['MISSION'] = 'TESS DVT'
    lc.meta['FLUX_ORIGIN'] = "LC_INIT"
    lc.meta['EXTNAME'] = hdu.header['EXTNAME']

    return lc


def has_ephem(lc):
    return ('TPERIOD' in lc.meta and
            'TEPOCH' in lc.meta and
            'TUNIT1' in lc.meta)


@loader_importer_registry('Light Curve')
class LightCurveImporter(BaseImporterToDataCollection):
    template_file = __file__, "./lightcurve.vue"

    create_ephemeris_available = Bool(False).tag(sync=True)
    create_ephemeris = Bool(True).tag(sync=True)

    # HDUList-specific options
    input_hdulist = Bool(False).tag(sync=True)
    extension_items = List().tag(sync=True)
    extension_selected = Any().tag(sync=True)
    extension_multiselect = Bool(True).tag(sync=True)

    icon_radialtocheck = Unicode(read_icon(os.path.join(ICON_DIR, 'radialtocheck.svg'), 'svg+xml')).tag(sync=True)  # noqa
    icon_checktoradial = Unicode(read_icon(os.path.join(ICON_DIR, 'checktoradial.svg'), 'svg+xml')).tag(sync=True)  # noqa

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.is_valid:
            return

        self.input_hdulist = isinstance(self.input, fits.HDUList)
        if self.input_hdulist:
            # TODO: allow multiselect and select_all by default,
            # update __call__ logic below to loop and add each independently including ephemerides,
            # modify default data_label logic to inject data_label within loop (and inform user)
            self.extension = SelectFileExtensionComponent(self,
                                                          items='extension_items',
                                                          selected='extension_selected',
                                                          multiselect='extension_multiselect',
                                                          manual_options=self.input,
                                                          filters=[hdu_is_valid])
            self.extension.select_all()
            # NOTE: data_label_default handled by changes to extension_selected
        else:
            if self.input.meta.get('QUARTER', None) is None:
                self.data_label_default = self.input.meta.get('OBJECT', 'Light curve')
            else:
                self.data_label_default = f"{self.input.meta.get('OBJECT', 'Light curve')} [Q{self.input.meta.get('QUARTER')}]"  # noqa

    @property
    def user_api(self):
        expose = ['create_ephemeris']
        if isinstance(self.input, fits.HDUList):
            expose += ['extension']
        return ImporterUserApi(self, expose)

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'lcviz'):
            # NOTE: temporary during deconfig process
            return False
        return (isinstance(self.input, LightCurve) or
                (isinstance(self.input, fits.HDUList)
                 and len([hdu for hdu in self.input if hdu_is_valid(hdu)])))  # noqa

    @observe('extension_selected')
    def _extension_selected_changed(self, event={}):
        if not hasattr(self, 'extension') or not self.input_hdulist:
            return

        self._clear_cache('output')

        if self.extension_multiselect:
            self.data_label_default = self.input[0].header.get('OBJECT', 'Light curve')
            self.create_ephemeris_available = any([has_ephem(lc) for lc in self.output])
        else:
            self.data_label_default = f"{self.input[0].header.get('OBJECT', 'Light curve')} [{self.extension.selected_item['name']}]"  # noqa
            self.create_ephemeris_available = has_ephem(self.output)

    @staticmethod
    def _get_supported_viewers():
        return [{'label': 'flux-vs-time', 'reference': 'lcviz-time-viewer'}]

    @cached_property
    def output(self):
        if isinstance(self.input, LightCurve):
            lc = self.input
        else:
            # HDUList case
            pri_header = self.input[0].header
            if self.extension_multiselect:
                lc = [hdulist_to_lightcurve(pri_header, hdu) for hdu in self.extension.selected_hdu]
            else:
                lc = hdulist_to_lightcurve(pri_header, self.extension.selected_hdu)

        flux_origin = lc.meta.get('FLUX_ORIGIN', None)  # i.e. PDCSAP or SAP
        if flux_origin == 'flux' or (flux_origin is None and 'flux' in getattr(lc, 'columns', [])):  # noqa
            # then make a copy of this column so it won't be lost when changing with the flux_column
            # plugin
            lc['flux:orig'] = lc['flux']
            if 'flux_err' in lc.columns:
                lc['flux:orig_err'] = lc['flux_err']
            lc.meta['FLUX_ORIGIN'] = 'flux:orig'

        return lc

    def add_to_data_collection(self, data, *args, **kwargs):
        lc_cls = data.__class__
        kwargs.setdefault('cls', lc_cls)
        data = _data_with_reftime(self.app, data)
        super().add_to_data_collection(data, *args, **kwargs)

    def __call__(self):
        if self.input_hdulist and self.extension_multiselect:
            data_label = self.data_label_value
            lcs = self.output
            with self.app._jdaviz_helper.batch_load():
                for lc, ext in zip(lcs, self.extension.selected_name):
                    self.add_to_data_collection(lc, f"{data_label} [{ext}]")
        else:
            super().__call__()
            lcs = [self.output]

        if self.create_ephemeris_available and self.create_ephemeris \
                and 'Ephemeris' in self.app._jdaviz_helper.plugins:
            for lc, ext in zip(lcs, self.extension.selected_name):
                if not has_ephem(lc):
                    continue
                ephem = self.app._jdaviz_helper.plugins['Ephemeris']
                ephem_component = self.app.return_unique_name(ext, ephem.component.choices)

                time_offset = int(lc.meta.get('TUNIT1').split('- ')[1].split(',')[0])
                period = lc.meta.get('TPERIOD', 1.0)
                t0 = lc.meta.get('TEPOCH', None) + time_offset - ephem._obj.reference_time

                ephem.add_component(ephem_component, set_as_selected=False)
                ephem.update_ephemeris(ephem_component, t0=t0, period=period, wrap_at=0.5)
                ephem.create_phase_viewer(ephem_component)



from glue.config import data_translator
from glue.core import Data, Subset
from ipyvue import watch
import warnings

import os
from glue.core.coordinates import Coordinates
from glue.core.component_id import ComponentID
import numpy as np
from scipy.interpolate import interp1d

from lightkurve import (
    LightCurve, KeplerLightCurve, TessLightCurve, FoldedLightCurve
)
from lightkurve.targetpixelfile import (
    KeplerTargetPixelFile, TessTargetPixelFile,
    TargetPixelFile, TargetPixelFileFactory
)
from lightkurve.utils import KeplerQualityFlags, TessQualityFlags

from astropy import units as u
from astropy.table import QTable
from astropy.time import Time
from astropy.wcs.wcsapi.wrappers.base import BaseWCSWrapper
from astropy.wcs.wcsapi import HighLevelWCSMixin

__all__ = ['TimeCoordinates', 'LightCurveHandler',
           'phase_comp_lbl',
           'data_not_folded',
           'is_lc', 'is_tpf', 'is_not_tpf',
           'enable_hot_reloading']


component_ids = {'dt': ComponentID('dt')}


class TimeCoordinates(Coordinates):
    """
    This is a sub-class of Coordinates that is intended for a time axis
    given by a :class:`~astropy.time.Time` array.
    """

    def __init__(self, times, reference_time=None, unit=u.d):
        if not isinstance(times, Time):  # pragma: no cover
            raise TypeError('values should be a Time instance')
        self._index = np.arange(len(times))
        self._times = times
        self.unit = unit

        # convert to relative time units
        self.reference_time = reference_time
        if self.reference_time is None:
            self.reference_time = times[0]
        self._values = (times - self.reference_time).to(unit)

        super().__init__(n_dim=1)

    @property
    def time_axis(self):
        return self._times

    def world_to_pixel_values(self, *world):
        if len(world) > 1:  # pragma: no cover
            raise ValueError('TimeCoordinates is a 1-d coordinate class '
                             'and only accepts a single scalar or array to convert')
        return interp1d(
            self._values.value, self._index, fill_value='extrapolate'
        )(world[0])

    def pixel_to_world_values(self, *pixel):
        if len(pixel) > 1:  # pragma: no cover
            raise ValueError('SpectralCoordinates is a 1-d coordinate class '
                             'and only accepts a single scalar or array to convert')
        return interp1d(
            self._index, self._values.value, fill_value='extrapolate'
        )(pixel[0])


class PaddedTimeWCS(BaseWCSWrapper, HighLevelWCSMixin):

    # Spectrum1D can use a 1D spectral WCS even for n-dimensional
    # datasets while glue always needs the dimensionality to match,
    # so this class pads the WCS so that it is n-dimensional.

    # NOTE: This class could be updated to use CompoundLowLevelWCS from NDCube.

    def __init__(self, wcs, times, ndim=3, reference_time=None, unit=u.d):
        self.temporal_wcs = TimeCoordinates(
            times, reference_time=reference_time, unit=unit
        )
        self.spatial_wcs = wcs
        self.flux_ndim = ndim
        self.spatial_keys = [f"spatial{i}" for i in range(0, self.flux_ndim-1)]

    @property
    def time_axis(self):
        return self.temporal_wcs.time_axis

    @property
    def pixel_n_dim(self):
        return self.flux_ndim

    @property
    def world_n_dim(self):
        return self.flux_ndim

    @property
    def world_axis_physical_types(self):
        return [self.temporal_wcs.world_axis_physical_types[0], *[None]*(self.flux_ndim-1)]

    @property
    def world_axis_units(self):
        return (self.temporal_wcs.world_axis_units[0], *[None]*(self.flux_ndim-1))

    def pixel_to_world_values(self, *pixel_arrays):
        # The ravel and reshape are needed because of
        # https://github.com/astropy/astropy/issues/12154
        px = np.array(pixel_arrays[0])
        world_arrays = [self.temporal_wcs.pixel_to_world_values(px.ravel()).reshape(px.shape),
                        *pixel_arrays[1:]]
        return tuple(world_arrays)

    def world_to_pixel_values(self, *world_arrays):
        # The ravel and reshape are needed because of
        # https://github.com/astropy/astropy/issues/12154
        wx = np.array(world_arrays[0])
        pixel_arrays = [self.temporal_wcs.world_to_pixel_values(wx.ravel()).reshape(wx.shape),
                        *world_arrays[1:]]
        return tuple(pixel_arrays)

    @property
    def world_axis_object_components(self):
        return [self.temporal_wcs.world_axis_object_components[0],
                *[(key, 'value', 'value') for key in self.spatial_keys]]

    @property
    def world_axis_object_classes(self):
        spectral_key = self.temporal_wcs.world_axis_object_components[0][0]
        obj_classes = {spectral_key: self.temporal_wcs.world_axis_object_classes[spectral_key]}
        for key in self.spatial_keys:
            obj_classes[key] = (u.Quantity, (), {'unit': u.pixel})

        return obj_classes

    @property
    def pixel_shape(self):
        return None

    @property
    def pixel_bounds(self):
        return None

    @property
    def pixel_axis_names(self):
        return tuple([self.temporal_wcs.pixel_axis_names[0], *self.spatial_keys])

    @property
    def world_axis_names(self):
        if self.flux_ndim == 2:
            names = ['Offset']
        else:
            names = [f"Offset{i}" for i in range(0, self.flux_ndim-1)]

        return ({}.get(self.temporal_wcs.world_axis_physical_types[0], ''),
                *names)

    @property
    def axis_correlation_matrix(self):
        return np.identity(self.flux_ndim).astype('bool')

    @property
    def serialized_classes(self):
        return False


@data_translator(LightCurve)
class LightCurveHandler:

    def to_data(self, obj, reference_time=None):
        is_folded = isinstance(obj, FoldedLightCurve)
        time = obj.time_original if is_folded and hasattr(obj, 'time_original') else obj.time
        time_coord = TimeCoordinates(time, reference_time=reference_time)
        data = Data(coords=time_coord)

        if hasattr(obj, 'label'):
            data.label = obj.label

        data.meta.update(obj.meta)
        data.meta.update(
            {"reference_time": time_coord.reference_time}
        )
        data[component_ids['dt']] = (obj.time - time_coord.reference_time).to(time_coord.unit)
        data.get_component('dt').units = str(time_coord.unit)

        # LightCurve is a subclass of astropy TimeSeries, so
        # collect all other columns in the TimeSeries:
        for component_label in obj.colnames:

            component_data = getattr(obj, component_label)
            if is_folded and component_label == 'time':
                ephem_comp = obj.meta.get('_LCVIZ_EPHEMERIS', {}).get('ephemeris')
                if ephem_comp is None:
                    continue
                component_label = f'phase:{ephem_comp}'

            if component_label not in component_ids:
                component_ids[component_label] = ComponentID(component_label)
            cid = component_ids[component_label]

            data[cid] = component_data
            if hasattr(component_data, 'unit'):
                try:
                    data.get_component(cid).units = str(component_data.unit)
                except KeyError:  # pragma: no cover
                    continue

        data.meta.update({'uncertainty_type': 'std'})

        # if the anticipated x and y axes are the first two components in the
        # Data object, the viewer will load those components correctly before
        # you hit the call to `viewer.set_plot_axes`:
        reordered_components = {comp.label: comp for comp in data.components}
        dt_comp = reordered_components.pop('dt')
        flux_comp = reordered_components.pop('flux')
        data.reorder_components(
            [dt_comp, flux_comp] +
            list(reordered_components.values())
        )

        return data

    def to_object(self, data_or_subset):
        """
        Convert a glue Data object to a lightkurve.LightCurve object.

        Parameters
        ----------
        data_or_subset : `glue.core.data.Data` or `glue.core.subset.Subset`
            The data to convert to a LightCurve object
        attribute : `glue.core.component_id.ComponentID`
            The attribute to use for the LightCurve data
        """

        if isinstance(data_or_subset, Subset):
            data = data_or_subset.data
            subset_state = data_or_subset.subset_state
        else:
            data = data_or_subset
            subset_state = None

        # Copy over metadata
        kwargs = {'meta': data.meta.copy()}

        # extract a Time object out of the TimeCoordinates object:
        time = data.coords.time_axis

        if subset_state is None:
            # pass through mask of all True's if no glue subset is chosen
            glue_mask = np.ones(len(time)).astype(bool)
        else:
            # get the subset mask from glue:
            glue_mask = data.get_mask(subset_state=subset_state)
            # apply the subset mask to the time array:
            time = time[glue_mask]

        columns = [time]
        names = ['time']

        component_ids = data.main_components

        # we already handled time separately above, and `dt` is only used internally
        # in LCviz, so let's skip those IDs below:
        skip_components = [id for id in component_ids if id.label in ['time', 'dt']]
        for skip_comp in skip_components:
            component_ids.remove(skip_comp)

        for component_id in component_ids:
            if component_id.label in names:
                # avoid duplicate column
                continue
            component = data.get_component(component_id)

            values = component.data[glue_mask]

            if len(values) and isinstance(values[0], Time):
                values = Time(values.base)
            elif hasattr(component, 'units') and component.units != "None":
                try:
                    values = u.Quantity(values, component.units)
                except TypeError:
                    if component.units != "":
                        raise
                    # values could have been an array of strings with units ""
                    values = values

            if component_id.label not in names:
                columns.append(values)
                names.append(component_id.label)

        table = QTable(columns, names=names, masked=True, copy=False)
        return LightCurve(table, **kwargs)


class TPFHandler:
    quality_flag_cls = None
    tpf_attrs = ['flux', 'flux_bkg', 'flux_bkg_err', 'flux_err']
    meta_attrs = [
        'cadenceno',
        'campaign',
        'channel',
        'column',
        'dec',
        'hdu',
        'mission',
        'module',
        'nan_time_mask',
        'obsmode',
        'output',
        'pipeline_mask',
        'pos_corr1',
        'pos_corr2',
        'quality',
        'quarter',
        'ra',
        'row',
        'shape',
        'wcs'
    ]

    def to_data(self, obj, reference_time=None, unit=u.d):
        coords = PaddedTimeWCS(obj.wcs, obj.time, reference_time=reference_time, unit=unit)
        data = Data(coords=coords)

        flux_shape = obj.flux.shape

        if hasattr(obj, 'label'):
            data.label = obj.label

        data.meta.update(obj.meta)
        data.meta.update(
            {"reference_time": coords.temporal_wcs.reference_time}
        )

        data[component_ids['dt']] = np.broadcast_to(
            (
                obj.time - coords.temporal_wcs.reference_time
            ).to(coords.temporal_wcs.unit)[:, None, None], flux_shape
        )
        data.get_component('dt').units = str(coords.temporal_wcs.unit)

        # LightCurve is a subclass of astropy TimeSeries, so
        # collect all other columns in the TimeSeries:
        for component_label in self.tpf_attrs:

            component_data = getattr(obj, component_label)
            if component_label not in component_ids:
                component_ids[component_label] = ComponentID(component_label)
            cid = component_ids[component_label]

            data[cid] = component_data
            if hasattr(component_data, 'unit'):
                try:
                    data.get_component(cid).units = str(component_data.unit)
                except KeyError:  # pragma: no cover
                    continue

        data.meta.update({'uncertainty_type': 'std'})

        for attr in self.meta_attrs:
            value = getattr(obj, attr, None)
            data.meta.update({attr: value})

        # if the anticipated x and y axes are the first two components in the
        # Data object, the viewer will load those components correctly before
        # you hit the call to `viewer.set_plot_axes`:
        reordered_components = {comp.label: comp for comp in data.components}
        dt_comp = reordered_components.pop('dt')
        flux_comp = reordered_components.pop('flux')
        data.reorder_components(
            [dt_comp, flux_comp] +
            list(reordered_components.values())
        )

        return data

    def to_object(self, data_or_subset):
        """
        Convert a glue Data object to a lightkurve.KeplerTargetPixelFile object.

        Parameters
        ----------
        data_or_subset : `glue.core.data.Data` or `glue.core.subset.Subset`
            The data to convert to a KeplerTargetPixelFile object
        attribute : `glue.core.component_id.ComponentID`
            The attribute to use for the KeplerTargetPixelFile data
        """

        if isinstance(data_or_subset, Subset):
            data = data_or_subset.data
            subset_state = data_or_subset.subset_state
        else:
            data = data_or_subset
            subset_state = None

        # Copy over metadata

        meta = data.meta.copy()
        for attr in self.meta_attrs:
            # these attrs don't belong in the lightkurve object's meta:
            meta.pop(attr)

        # extract a Time object out of the TimeCoordinates object:
        time = data.coords.time_axis

        if subset_state is None:
            # pass through mask of all True's if no glue subset is chosen
            glue_mask = None
        else:
            # get the subset mask from glue:
            glue_mask = data.get_mask(subset_state=subset_state)
            # apply the subset mask to the time array:
            time = time[glue_mask]

        attrs_to_save = {'meta': meta, 'time': time}

        component_ids = data.main_components

        # we already handled time separately above, and `dt` is only used internally
        # in LCviz, so let's skip those IDs below:
        skip_components = [id for id in component_ids if id.label in ['time', 'dt']]
        for skip_comp in skip_components:
            component_ids.remove(skip_comp)

        for component_id in component_ids:
            if component_id.label in attrs_to_save:
                # avoid duplicate column
                continue
            component = data.get_component(component_id)
            values = component.data
            if glue_mask is not None:
                values = values[glue_mask]

            if component_id.label not in attrs_to_save:
                attrs_to_save[component_id.label] = values

        tpf_factory = TargetPixelFileFactory(*data.shape)

        for attr, values in attrs_to_save.items():
            if attr == 'time':
                values = values.value
            setattr(tpf_factory, attr, values)

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore',
                                    message='Could not detect filetype as TESSTargetPixelFile or KeplerTargetPixelFile, returning generic TargetPixelFile instead.')  # noqa
            tpf = tpf_factory.get_tpf()

        for attr in self.meta_attrs:
            # if this attribute exists and can be set:
            if hasattr(tpf, attr) and getattr(getattr(tpf, attr), 'fset', None) is not None:
                setattr(tpf, attr, data.meta[attr])

        tpf.quality_mask = self.quality_flag_cls.create_quality_mask(
            quality_array=tpf.hdu[1].data["QUALITY"], bitmask=tpf.quality_bitmask
        )

        return tpf


def enable_hot_reloading(watch_jdaviz=True):
    """
    Use ``watchdog`` to perform hot reloading.

    Parameters
    ----------
    watch_jdaviz : bool
        Whether to also watch changes to jdaviz upstream.
    """
    try:
        watch(os.path.dirname(__file__))
        if watch_jdaviz:
            import jdaviz
            watch(os.path.dirname(jdaviz.__file__))
    except ModuleNotFoundError:
        print((
            'Watchdog module, needed for hot reloading, not found.'
            ' Please install with `pip install watchdog`'))


@data_translator(KeplerLightCurve)
class KeplerLightCurveHandler(LightCurveHandler):
    # Works the same as LightCurve
    pass


@data_translator(TessLightCurve)
class TessLightCurveHandler(LightCurveHandler):
    # Works the same as LightCurve
    pass


@data_translator(KeplerTargetPixelFile)
class KeplerTPFHandler(TPFHandler):
    quality_flag_cls = KeplerQualityFlags


@data_translator(TessTargetPixelFile)
class TessTPFHandler(TPFHandler):
    quality_flag_cls = TessQualityFlags


def phase_comp_lbl(component):
    return f'phase:{component}'


# plugin component filters
def data_not_folded(data):
    return data.meta.get('_LCVIZ_EPHEMERIS', None) is None


def is_lc(data):
    return (len(data.shape) == 1
            and isinstance(data.coords, TimeCoordinates))


def is_tpf(data):
    return len(data.shape) == 3


def is_not_tpf(data):
    return not is_tpf(data)


lightkurve_handlers = {
    LightCurve: LightCurveHandler(),
    TargetPixelFile: TPFHandler(),
}


def _data_with_reftime(app, light_curve):
    # grab the first-found reference time in the data collection:
    ff_reference_time = None
    for existing_data in app.data_collection:
        if hasattr(existing_data, 'meta') and 'reference_time' in existing_data.meta:
            ff_reference_time = existing_data.meta.get('reference_time', None)
            if ff_reference_time is not None:
                break

    if isinstance(light_curve, Data):
        light_curve.meta['reference_time'] = ff_reference_time
        return light_curve

    # convert to glue Data manually, so we may edit the `dt` component if necessary:
    for expected_cls, handler in lightkurve_handlers.items():
        if isinstance(light_curve, expected_cls):
            return handler.to_data(light_curve, reference_time=ff_reference_time)
    else:
        raise ValueError(f"No handler found for {light_curve} of type {type(light_curve)}")



import numpy as np
from traitlets import observe

from jdaviz.configs.default.plugins import Markers
from jdaviz.core.registries import tray_registry

__all__ = ['Markers']


@tray_registry('g-markers', label="Markers",
               category='core', sidebar='viewers', subtab=1, overwrite=True)
class Markers(Markers):
    """
    See the :ref:`Markers Plugin Documentation <markers>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * :meth:`clear_table`
    * :meth:`~jdaviz.core.template_mixin.TableMixin.export_table`
    """
    _default_table_values = {'time': np.nan,
                             'time:unit': '',
                             'phase': np.nan,
                             'ephemeris': '',
                             'pixel': (np.nan, np.nan),
                             'value': np.nan,
                             'value:unit': ''}

    def __init__(self, *args, **kwargs):
        kwargs['headers'] = ['time', 'time:unit', 'phase', 'ephemeris',
                             'pixel', 'value', 'value:unit', 'viewer']
        super().__init__(*args, **kwargs)

    @observe('vdocs')
    def _update_docs_link(self, *args):
        self.docs_link = f"https://lcviz.readthedocs.io/en/{self.vdocs}/plugins.html#markers"

    @property
    def coords_info(self):
        return self.app.session.application._tools['g-coords-info']



from traitlets import observe

from jdaviz.core.viewer_creators import BaseViewerCreator
from jdaviz.core.registries import viewer_creator_registry
from jdaviz.core.user_api import ViewerCreatorUserApi

from lcviz.components import EphemerisSelectMixin
from lcviz.viewers import TimeScatterView, PhaseScatterView
from lcviz.utils import is_lc

__all__ = ['FluxVsTimeViewerCreator']


@viewer_creator_registry('Flux vs Time', overwrite=True)
class FluxVsTimeViewerCreator(BaseViewerCreator):
    template_file = __file__, "../base_viewer_creator.vue"

    def __init__(self, app, **kwargs):
        super().__init__(app, **kwargs)
        self.dataset.filters = [is_lc]
        self.viewer_label_default = 'flux-vs-time'  # TODO: append suffix if taken

    @property
    def viewer_class(self):
        return TimeScatterView


@viewer_creator_registry('Flux vs Phase', overwrite=True)
class FluxVsPhaseViewerCreator(BaseViewerCreator, EphemerisSelectMixin):
    template_file = __file__, "./phase_viewer_creator.vue"

    def __init__(self, app, **kwargs):
        super().__init__(app, **kwargs)
        self.dataset.filters = [is_lc]
        self.viewer_label_default = 'flux-vs-phase'  # TODO: append suffix if taken

    @property
    def viewer_class(self):
        return PhaseScatterView

    @property
    def user_api(self):
        return ViewerCreatorUserApi(self, expose=['ephemeris'])

    @observe('ephemeris_items')
    def _update_ephemeris_items(self, change):
        self._set_is_relevant()
        if not self.ephemeris_selected:
            self.ephemeris.select_default()

    @observe('ephemeris_selected')
    def _ephemeris_selected_changed(self, change):
        self.viewer_label_default = f'flux-vs-phase:{self.ephemeris.selected}'

    def _set_is_relevant(self):
        if not len(self.dataset_items) or not len(self.ephemeris_items):
            self.is_relevant = False
        else:
            self.is_relevant = True

    def __call__(self):
        nv = super().__call__()
        ephem_plg = self.app._jdaviz_helper.plugins['Ephemeris']
        ephem_plg._obj._set_viewer_to_ephem_component(nv._obj, self.ephemeris.selected)
        return nv



# Licensed under a 3-clause BSD style license - see LICENSE.rst

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''

from . import utils  # noqa
from .utils import enable_hot_reloading  # noqa: F401
from . import state  # noqa

# Expose subpackage API at package level.
from .plugins import *  # noqa
from .tools import *  # noqa
from .viewers import *  # noqa
from .helper import *  # noqa

from .loaders import *  # noqa
from .viewer_creators import *  # noqa

from jdaviz import gca

jdaviz_app = gca()
# inject loaders/plugins into the jdaviz deconfigged app
jdaviz_app.app.update_tray_items_from_registry()
jdaviz_app.app.update_loaders_from_registry()
jdaviz_app.app.update_new_viewers_from_registry()

# redirect top-level calls to the deconfigged jdaviz app
_expose = ['show', 'load', 'batch_load',
           'toggle_api_hints',
           'plugins',
           'loaders',
           'viewers']
_incl = ['App', 'enable_hot_reloading', '__version__']
_temporary_incl = ['LCviz']
__all__ = _expose + _incl + _temporary_incl


def __dir__():
    return sorted(__all__)



import pytest
import numpy as np
from glue.core.roi import XRangeROI, YRangeROI
from astropy.time import Time
from astropy.utils.data import download_file
from lightkurve import (LightCurve, KeplerTargetPixelFile,
                        search_lightcurve, search_targetpixelfile)
from lightkurve.io import kepler
import astropy.units as u

from lcviz.utils import TimeCoordinates


@pytest.mark.remote_data
def test_kepler_via_mast_local_file(helper):
    url = (
        'https://archive.stsci.edu/pub/kepler/'
        'lightcurves/0014/001429092/kplr001429092-2009166043257_llc.fits'
    )  # 188 KB

    path = download_file(url, cache=True, timeout=100)
    helper.load_data(path)

    data = helper.app.data_collection[0]
    flux_arr = data['flux']
    flux_unit = u.Unit(data.get_component('flux').units)
    flux = flux_arr * flux_unit

    assert isinstance(data.coords, TimeCoordinates)
    assert isinstance(flux, u.Quantity)
    assert flux.unit.is_equivalent(u.electron / u.s)


@pytest.mark.remote_data
def test_kepler_via_mast_preparsed(helper):
    url = (
     'https://archive.stsci.edu/pub/kepler/'
     'lightcurves/0014/001429092/kplr001429092-2009166043257_llc.fits'
    )  # 188 KB

    light_curve = kepler.read_kepler_lightcurve(url)
    helper.load_data(light_curve)

    data = helper.app.data_collection[0]
    flux_arr = data['flux']
    flux_unit = u.Unit(data.get_component('flux').units)
    flux = flux_arr * flux_unit

    assert isinstance(data.coords, TimeCoordinates)
    assert isinstance(flux, u.Quantity)
    assert flux.unit.is_equivalent(u.electron / u.s)


@pytest.mark.remote_data
def test_kepler_tpf_via_lightkurve(helper):
    tpf = search_targetpixelfile("KIC 001429092",
                                 mission="Kepler",
                                 cadence="long",
                                 quarter=10).download()
    helper.load_data(tpf)
    assert helper.get_data().shape == (4447, 4, 6)  # (time, x, y)
    assert helper.app.data_collection[0].get_object(cls=KeplerTargetPixelFile).shape == (4447, 4, 6)


@pytest.mark.remote_data
def test_mult_lc_reftime(helper):
    lc1 = search_lightcurve("HAT-P-11", mission="Kepler",
                            cadence="long", quarter=9).download()
    lc2 = search_lightcurve("HAT-P-11", mission="Kepler",
                            cadence="long", quarter=10).download()
    helper.load_data(lc1, data_label='Q9')
    helper.load_data(lc2, data_label='Q10')
    assert helper.app.data_collection[0].meta.get('reference_time') == helper.app.data_collection[1].meta.get('reference_time')  # noqa


def test_synthetic_lc(helper):
    time = Time(np.linspace(2460050, 2460060), format='jd')
    flux = np.ones(len(time)) * u.electron / u.s
    flux_err = 0.1 * np.ones_like(flux)
    lc = LightCurve(time=time, flux=flux, flux_err=flux_err)
    helper.load_data(lc)

    data = helper.app.data_collection[0]
    flux_arr = data['flux']
    flux_unit = u.Unit(data.get_component('flux').units)
    flux = flux_arr * flux_unit

    assert isinstance(data.coords, TimeCoordinates)
    assert isinstance(flux, u.Quantity)
    assert flux.unit.is_equivalent(u.electron / u.s)


def test_apply_xrangerois(helper, light_curve_like_kepler_quarter):
    lc = light_curve_like_kepler_quarter
    helper.load_data(lc)
    viewer = helper.default_time_viewer._obj
    subset_plugin = helper.plugins['Subset Tools']

    # the min/max of temporal regions can be defined in two ways:
    time_ranges = [
        [6, 8],  # in same units as the x axis, OR
        Time(['2011-07-19', '2011-07-23'])  # directly with a Time object
    ]

    for time_range in time_ranges:
        subset_plugin._obj.subset_selected = "Create New"
        viewer.apply_roi(XRangeROI(*time_range))

    subsets = helper.app.get_subsets()

    subset_1_bounds_jd = subsets['Subset 1'][0]['region'].jd
    subset_2_bounds_jd = subsets['Subset 2'][0]['region'].jd

    np.testing.assert_allclose(subset_1_bounds_jd, [2455745., 2455747.])
    np.testing.assert_allclose(subset_2_bounds_jd, [2455761.50076602, 2455765.50076602])


def test_apply_yrangerois(helper, light_curve_like_kepler_quarter):
    lc = light_curve_like_kepler_quarter
    helper.load_data(lc)
    viewer = helper.default_time_viewer._obj
    subset_plugin = helper.plugins['Subset Tools']

    subset_plugin._obj.subset_selected = "Create New"
    viewer.apply_roi(YRangeROI(1, 1.05))

    subsets = helper.app.get_subsets()

    # TODO: subsets['Subset 1'][0]['region'] is still returning a Time object

    subset_state = subsets['Subset 1'][0]['subset_state']

    np.testing.assert_allclose([subset_state.lo, subset_state.hi], [1, 1.05])


def test_data_label(helper, light_curve_like_kepler_quarter):
    # add data without specifying data label:
    helper.load_data(light_curve_like_kepler_quarter)
    object_name = helper.app.data_collection[-1].meta['OBJECT']
    assert helper.app.data_collection[-1].label == f'{object_name} [Q10]'

    # specify label, check that quarter isn't appended:
    data_label = 'Cool target'
    helper.load_data(light_curve_like_kepler_quarter, data_label=data_label)
    assert helper.app.data_collection[-1].label == data_label



from astropy.time import Time
from traitlets import Bool, observe
from glue.config import data_translator

from jdaviz.core.custom_traitlets import IntHandleEmpty
from jdaviz.core.events import (ViewerAddedMessage, ViewerRemovedMessage)
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        DatasetSelectMixin, AddResultsMixin,
                                        skip_if_no_updates_since_last_active,
                                        with_spinner, with_temp_disable)
from jdaviz.core.user_api import PluginUserApi

from lcviz.components import FluxColumnSelectMixin
from lcviz.events import EphemerisChangedMessage
from lcviz.marks import LivePreviewBinning
from lcviz.viewers import TimeScatterView, PhaseScatterView
from lcviz.components import EphemerisSelectAllowNoneMixin
from lcviz.utils import is_lc, phase_comp_lbl, _data_with_reftime


__all__ = ['Binning']


@tray_registry('binning', label="Binning", category='data:manipulation')
class Binning(PluginTemplateMixin, FluxColumnSelectMixin, DatasetSelectMixin,
              EphemerisSelectAllowNoneMixin, AddResultsMixin):
    """
    See the :ref:`Binning Plugin Documentation <binning>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Dataset to bin.
    * ``ephemeris`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      Label of the component corresponding to the active ephemeris.
    * :meth:`input_lc`
      Data used as input to binning, based on ``dataset`` and ``ephemeris``.
    * ``n_bins`` : int
    * ``add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * :meth:`bin`
    """
    template_file = __file__, "binning.vue"
    uses_active_status = Bool(True).tag(sync=True)

    show_live_preview = Bool(True).tag(sync=True)

    n_bins = IntHandleEmpty(100).tag(sync=True)
    bin_enabled = Bool(True).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._plugin_description = 'Bin input light curve.'

        self._set_results_viewer()

        # TODO: replace with add_filter('not_from_this_plugin') if upstream PR accepted/released
        # https://github.com/spacetelescope/jdaviz/pull/2239
        def not_from_binning_plugin(data):
            return data.meta.get('Plugin', None) != self.__class__.__name__
        self.dataset.add_filter(not_from_binning_plugin, is_lc)

        # TODO: viewer added also needs to repopulate marks
        self.hub.subscribe(self, ViewerAddedMessage, handler=self._on_add_viewer)
        self.hub.subscribe(self, ViewerRemovedMessage, handler=self._set_results_viewer)
        self.hub.subscribe(self, EphemerisChangedMessage, handler=self._on_ephemeris_update)

        self._set_relevant()

    @observe('dataset_items')
    def _set_relevant(self, *args):
        if not len(self.dataset_items):
            self.irrelevant_msg = 'No valid datasets loaded'
        else:
            self.irrelevant_msg = ''

    @property
    def user_api(self):
        expose = ['show_live_preview', 'dataset', 'ephemeris', 'input_lc',
                  'n_bins', 'add_results', 'bin']
        return PluginUserApi(self, expose=expose)

    @property
    def ephemeris_plugin(self):
        return self.ephemeris.ephemeris_plugin

    @property
    def ephemeris_dict(self):
        if self.ephemeris_selected == 'No ephemeris':
            return {}
        return self.ephemeris_plugin.ephemerides.get(self.ephemeris_selected)

    @property
    def input_lc(self):
        return self.ephemeris.get_data_for_dataset(self.dataset)

    @property
    def marks(self):
        marks = {}
        for viewer in self.app._viewer_store.values():
            for mark in viewer.figure.marks:
                if isinstance(mark, LivePreviewBinning):
                    marks[viewer.reference] = mark
                    break
            else:
                mark = LivePreviewBinning(viewer, visible=self.is_active)
                viewer.figure.marks = viewer.figure.marks + [mark]
                marks[viewer.reference] = mark
        return marks

    def _clear_marks(self):
        for mark in self.marks.values():
            if mark.visible:
                mark.clear()
                mark.visible = False

    @observe("dataset_selected", "ephemeris_selected")
    def _set_default_results_label(self, event={}):
        '''Generate a label and set the results field to that value'''
        if not hasattr(self, 'ephemeris'):
            return
        label = f"binned {self.dataset_selected}"
        if self.ephemeris_selected not in self.ephemeris._manual_options:
            label += f":{self.ephemeris_selected}"
        self.results_label_default = label

    @observe("ephemeris_selected")
    def _set_results_viewer(self, event={}):
        if not hasattr(self, 'ephemeris'):
            return

        def viewer_filter(viewer):
            if self.ephemeris_selected in self.ephemeris._manual_options:
                return isinstance(viewer, TimeScatterView)
            if not isinstance(viewer, PhaseScatterView):
                # ephemeris selected, but no active phase viewers
                return False
            return viewer._ephemeris_component == self.ephemeris_selected

        self.add_results.viewer.filters = [viewer_filter]

    def _on_add_viewer(self, msg):
        self._set_results_viewer()
        self._live_update()

    @observe('is_active', 'show_live_preview')
    def _toggle_marks(self, event={}):
        visible = self.show_live_preview and self.is_active

        for viewer_ref, mark in self.marks.items():
            if not visible:
                this_visible = False
            elif self.ephemeris_selected == 'No ephemeris':
                this_visible = True
            else:
                viewer = self.app.get_viewer(viewer_ref)
                viewer_ephem = getattr(viewer, '_ephemeris_component', None)
                this_visible = viewer_ephem == self.ephemeris_selected

            mark.visible = this_visible

        if visible and event.get('name', '') in ('is_active', 'show_live_preview'):
            # then the marks themselves need to be updated
            self._live_update(event)

    @observe('flux_column_selected', 'dataset_selected',
             'ephemeris_selected',
             'n_bins', 'previews_temp_disabled')
    @skip_if_no_updates_since_last_active()
    @with_temp_disable(timeout=0.3)
    def _live_update(self, event={}):
        self.bin_enabled = self.n_bins != '' and self.n_bins > 0

        if not self.show_live_preview or not self.is_active or not self.bin_enabled:
            self._clear_marks()
            return

        if event.get('name', '') not in ('is_active', 'show_live_preview'):
            # mark visibility hasn't been handled yet
            self._toggle_marks()

        try:
            lc = self.bin(add_data=False)
        except Exception:
            self._clear_marks()
            self.bin_enabled = False
            return
        else:
            self.bin_enabled = True

        # TODO: remove the need for this (inconsistent quantity vs value setting in lc object)
        lc_time = getattr(lc.time, 'value', lc.time)

        if self.ephemeris_selected == 'No ephemeris':
            ref_time = lc.meta.get('reference_time', 0)
            ref_time = getattr(ref_time, 'value', ref_time)
            times = lc_time - ref_time
        else:
            times = lc_time

        for viewer_id, mark in self.marks.items():
            if self.ephemeris_selected == 'No ephemeris':
                # TODO: fix this to be general and not rely on ugly id
                do_phase = viewer_id != 'lcviz-0'
            else:
                do_phase = False

            if do_phase:
                mark.update_ty(times, lc.flux.value)
            else:
                mark.times = []
                mark.update_xy(times, lc.flux.value)

    def _on_ephemeris_update(self, msg):
        if not self.show_live_preview or not self.is_active:
            return

        if msg.ephemeris_label != self.ephemeris_selected:
            return

        self._live_update()

    @with_spinner()
    def bin(self, add_data=True):
        if self.n_bins == '' or self.n_bins <= 0:
            raise ValueError("n_bins must be a positive integer")

        input_lc = self.input_lc

        # fix for bug in py310, see:
        # https://github.com/spacetelescope/lcviz/pull/194
        if "NORMALIZE_PHASE" not in input_lc.meta:
            input_lc.meta["NORMALIZE_PHASE"] = False

        lc = input_lc.bin(time_bin_size=(input_lc.time[-1]-input_lc.time[0]).value/self.n_bins)
        if self.ephemeris_selected != 'No ephemeris':
            # lc.time.value are actually phases, so convert to times starting at time t0
            times = self.ephemeris_plugin.phases_to_times(lc.time.value, self.ephemeris_selected)

            # set the time_original column, leaving the time column as phases
            time_col = Time(times,
                            format=input_lc.time_original.format,
                            scale=input_lc.time_original.scale)
            lc.add_column(time_col, name="time_original", index=len(lc._required_columns))

            lc.meta.update({'_LCVIZ_BINNED': True})

            # convert to glue Data manually, so we may edit the `phase` component:
            handler, _ = data_translator.get_handler_for(lc)
            data = handler.to_data(lc)

            # here we use the `value` attribute of `lc.time`, which has units of *phase*:
            self.app._jdaviz_helper._set_data_component(data,
                                                        phase_comp_lbl(self.ephemeris_selected),
                                                        lc.time.value)

        else:
            # need to send through parser-logic to assign the correct reference time
            data = _data_with_reftime(self.app, lc)

        if add_data:
            # add data to the collection/viewer
            # NOTE: lc will have _LCVIZ_EPHEMERIS set if phase-folded
            self._set_results_viewer()
            self.add_results.add_results_from_plugin(data or lc)

            if self.ephemeris_selected != 'No ephemeris':
                # prevent phase axis from becoming a time axis:
                ephemeris_plugin = self.app._jdaviz_helper.plugins['Ephemeris']
                phase_comp = self.app._jdaviz_helper._component_ids[phase_comp_lbl(self.ephemeris_selected)]  # noqa
                for pv in ephemeris_plugin._obj._get_phase_viewers(self.ephemeris_selected):
                    pv.state.x_att = phase_comp
                # by resetting x_att, the preview marks may have dissappeared
                self._live_update()

        return lc

    def vue_apply(self, event={}):
        self.bin(add_data=True)



from numpy.testing import assert_allclose

from lightkurve.periodogram import LombScarglePeriodogram, BoxLeastSquaresPeriodogram


def test_docs_snippets(helper, light_curve_like_kepler_quarter):
    lcviz, lc = helper, light_curve_like_kepler_quarter

    lcviz.load(lc)
    # lcviz.show()

    freq = lcviz.plugins['Frequency Analysis']
    freq.open_in_tray()
    freq.method = 'Lomb-Scargle'
    freq.xunit = 'period'
    periodogram = freq.periodogram
    print(periodogram)


def test_plugin_frequency_analysis(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)

    freq = helper.plugins['Frequency Analysis']
    freq.open_in_tray()

    assert freq.method == 'Lomb-Scargle'
    assert freq._obj.err == ''
    assert isinstance(freq.periodogram, LombScarglePeriodogram)

    freq.method = 'Box Least Squares'
    assert freq._obj.err == ''
    assert isinstance(freq.periodogram, BoxLeastSquaresPeriodogram)

    assert freq.xunit == 'frequency'
    assert freq._obj.plot.figure.axes[0].label == 'frequency (1 / d)'

    freq.xunit = 'period'
    assert freq._obj.plot.figure.axes[0].label == 'period (d)'
    line_x = freq._obj.plot.layers['periodogram'].layer['x']
    assert_allclose((line_x.min(), line_x.max()), (0.3508333334885538, 31.309906458683404))

    freq.auto_range = False
    assert_allclose((freq.minimum, freq.maximum), (1, 10))
    while freq._obj.spinner:
        pass
    line_x = freq._obj.plot.layers['periodogram'].layer['x']
    assert_allclose((line_x.min(), line_x.max()), (1, 10.00141))

    freq.xunit = 'frequency'
    assert_allclose((freq.minimum, freq.maximum), (0.1, 1))
    while freq._obj.spinner:
        pass
    line_x = freq._obj.plot.layers['periodogram'].layer['x']
    assert_allclose((line_x.min(), line_x.max()), (0.0999859, 1))
